#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-
# ----------------------------------------
# Copyright (C) 2010-2011  Frank Pählke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import sys
from toureditorlib.main import TourEditor, TourView
from toureditorlib.tourfile import TourFile

root = TourEditor()

for filename in sys.argv[1:]:
    TourView(TourFile.read (filename, 'iso-8859-1'))

root.mainloop()
root.destroy()



# -*- coding: utf-8 -*-
# ----------------------------------------
# Copyright (C) 2010-2011  Frank Pählke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import array
import math
from Tkinter import *
import tkFileDialog
from svgobject.svg import *
try:
    from svgobject.pdf import PDFWriter
    _no_pdf = False
except ImportError:
    _no_pdf = True

class TourGraph(Toplevel):

    def __init__ (self, parent):
        Toplevel.__init__ (self)
        self.parent = parent
        parent.registerListener (self)

        self.zoom = 1.0
        self.grid_rowconfigure (1, weight=1)
        self.grid_columnconfigure (0, weight=1)
        self.buttons = {}

        # button frame
        bf = Frame (self);
        bf.grid (row=0, column=0, columnspan=2, pady=1)

        self.buttons['close'] = b = Button (bf, text='Close', command=self.close)
        b.pack (side=LEFT)

        self.buttons['zoomIn'] = b = Button (bf, text="Zoom in", command=self.zoomIn)
        b.pack (side=LEFT)

        self.buttons['zoomOut'] = b = Button (bf, text="Zoom out", command=self.zoomOut)
        b.pack (side=LEFT)

        self.zoomlabel = Label (bf) # text will be set in self.repaint()
        self.zoomlabel.pack (side=LEFT)

        self.buttons['exportSVG'] = b = Button (bf, text="Export SVG", command=self.exportSVG)
        b.pack (side=LEFT)

        if not _no_pdf:
            self.buttons['exportPDF'] = b = Button (bf, text="Export PDF", command=self.exportPDF)
            b.pack (side=LEFT)
            
        self.canvas = Canvas (self, width=640, height=300,
                              borderwidth=2, relief=SUNKEN)
        self.canvas.grid (sticky=NSEW)

        self.scrollY = Scrollbar (self, command=self.canvas.yview)
        self.canvas.config (yscrollcommand=self.scrollY.set)
        self.scrollY.grid (row=1, column=1, sticky=NS)

        self.scrollX = Scrollbar(self, command=self.canvas.xview, orient=HORIZONTAL)
        self.canvas.config (xscrollcommand=self.scrollX.set)
        self.scrollX.grid (row=2, column=0, sticky=EW)

        self.refresh()

    def refresh (self):
        "repaint graph from tourfile data"
        tf = self.parent.tourfile       
        self.dscale = None #tf.dscale 
        self.hscale = None #tf.hscale
        self.sscale = None #tf.sscale
        self.title (tf.title)
        self.svg = SVG (640, 300, unit="px")
        self.drawSVG (self.svg)
        self.repaint()

    def repaint(self):
        "repaint canvas after zoom factor has changed"
        self.zoomlabel.config (text="Zoom: "+str(int(round(100*self.zoom)))+"%")
        self.canvas.delete (ALL)
        self.canvas.config (scrollregion=(0, 0, self.zoom*self.svg.width, self.zoom*self.svg.height))
        self.svg.paint (self.canvas, self.zoom)

    def zoomIn(self):
        self.zoom = self.zoom * math.sqrt(2)
        self.repaint()

    def zoomOut(self):
        self.zoom = self.zoom / math.sqrt(2)
        self.repaint()

    def close(self):
        self.parent.deregisterListener (self)
        self.destroy()

    def exportPDF(self):
        filename = tkFileDialog.asksaveasfilename (
            title='Export PDF', filetypes=[('PDF', '*.pdf')])
        if filename != None:
            tf = self.parent.tourfile
            svg = SVG (297, 210, unit="mm", view_box=(-51.25, -112.5, 742.5, 525))
            self.drawSVG (svg)
            writer = PDFWriter (filename, tf.title, tf.copyright)
            writer.writeSVG (svg)
            writer.close()

    def exportSVG(self):
        filename = tkFileDialog.asksaveasfilename (
            title='Export SVG',
            filetypes=[('SVG', '*.svg'), ('Compressed SVG', '*.svgz')])
        if filename != None:
            svg = SVG (640, 300, unit='pt')
            self.drawSVG (svg)
            if filename.lower().endswith ('z'):
                svg.writeSVG (filename, True)
            else:
                svg.writeSVG (filename, False)

    def drawSVG (self, svg):
        "draw graph into SVG object"
        tf = self.parent.tourfile
        if tf.is_smoothed:
            dist = tf.getcolumn ('smooth_distance')
            height = tf.getcolumn ('smooth_height')
        else:
            dist = tf.getcolumn ('distance')
            height = tf.getcolumn ('height')
 
        graph = Group (svg, style=Style (font_family="Arial,Helvetica,sans-serif", font_size="14"))

        # calculate slope
        slope = [ ((b-a)/(d-c)/10.0 if d != c else 0) for a,b,c,d in
                  zip(height[0:-1], height[1:], dist[0:-1], dist[1:]) ]

        # calculate scales
        d0,d1,dIntv = self.calcXScale(dist)
        h0,h1,hIntv,s0,s1,sIntv = self.calcYScale(height,slope)
        dTics = int((d1-d0)/dIntv)
        hTics = (h1-h0)/hIntv
        sTics = (s1-s0)/sIntv

        # background
        Rect (graph, 0, 0, 639, 299,
              style=Style (fill="#ffffff", stroke="#000000", stroke_width="1"))

        # title
        if tf.title != None:
            Text (graph, 320, 30, tf.title,
                  style=Style (text_anchor="middle", font_weight="bold", font_size="16"))
        # copyright
        if tf.copyright != None:
            Text (graph, 636, 296, tf.copyright,
                  style=Style (text_anchor="end", font_size="10", fill="#999999"))

        # canvas
        svgcanvas = Group (graph, translate=(80,50))
        Rect (svgcanvas, 0, 0, 500, 200, style=Style(fill="#cccccc"))

        # X minor grid
        if dTics <= 12:
            g = Group (svgcanvas, style=Style (stroke="#999999",stroke_width=1))
            for i in range (dTics):
                d = d0 + (i+0.5)*dIntv;
                x = int(round(self.scale(d, d0, d1, 0, 500)))
                Line (g, x, 0, x, 200)

        # Y minor grid
        if sTics <= 7:
            g = Group (svgcanvas, style=Style (stroke="#999999",stroke_width=1))
            for i in range (sTics):
                s = s0 + (i+0.5)*sIntv;
                y = int(round(self.scale(s, s0, s1, 200, 0)))
                Line (g, 0, y, 500, y)

        # X grid
        g = Group (svgcanvas, style=Style (stroke="#000000",stroke_width=1))
        for i in range (dTics+1):
            d = d0 + i*dIntv;
            x = int(round(self.scale(d, d0, d1, 0, 500)))
            Line (g, x, 0, x, 205)

        # X labels
        g = Group (svgcanvas, style=Style (text_anchor="middle"))
        Text (g, 250, 240, "Strecke [km]")
        for i in range (dTics+1):
            d = d0 + i*dIntv
            x = int(round(500 * float(d-d0) / float(d1-d0)))
            Text (g, x, 220, str(round(d,1)))

        # Y grid
        g = Group (svgcanvas, style=Style (stroke="#000000",stroke_width=1))
        for i in range (hTics+1):
            h = h0 + i*hIntv
            s = s0 + i*sIntv
            y = round(self.scale(h, h0, h1, 200, 0))
            if (s == 0) or (h == 0): # zero axis
                Line (g, -5, y, 505, y, style=Style(stroke_width=2))
            else:
                Line (g, -5, y, 505, y)

        # Y labels (primary)
        g = Group (svgcanvas, translate=(-55,100), rotate=-90)
        Text (g, 0, 0, u"blau: Höhe [m]", style=Style(text_anchor="middle"))
        g = Group (svgcanvas, style=Style(text_anchor="end"))
        for i in range (hTics+1):
            h = h0 + i*hIntv
            y = round(self.scale(h, h0, h1, 200, 0))
            Text (g, -10, y+5, str(int(round(h)))) # TODO: dy="0.5ex"

        # Y labels (secondary)
        g = Group (svgcanvas, translate=(545,100), rotate=-90)
        Text (g, 0, 0, "rot: Steigung [%]", style=Style(text_anchor="middle"))
        g = Group (svgcanvas, style=Style(text_anchor="start"))
        for i in range (sTics+1):
            s = s0 + i*sIntv
            y = round(self.scale(s, s0, s1, 200, 0))
            Text (g, 510, y+5, str(int(round(s)))) # TODO: dy="0.5ex"

        # slope curve (first, i.e. behind height curve)
        dist2 = map((lambda x,y: (x+y)/2), dist[0:-1], dist[1:])
        self.drawCurve (svgcanvas, dist2, d0, d1, slope, s0, s1, "#ff0000")

        # height curve (second, i.e. before slope curve)
        self.drawCurve (svgcanvas, dist, d0, d1, height, h0, h1, "#0000ff")

    # calculate scaling parameters X axis
    # returns: dBottom, dTop, dIntv

    def calcXScale (self, dist):
        if self.dscale != None:
            (dMin, dMax) = string.split(self.dscale, ',')
            dMin = float(dMin)
            dMax = float(dMax)
        else:
            dMin = min(dist)
            dMax = max(dist)

        # determine scale
        for dIntv in [0.1, 0.2, 0.5, 1, 2, 4, 5]:
            # round up/down
            dBottom = dIntv * math.floor(dMin/dIntv)
            dTop = dIntv * math.ceil(dMax/dIntv)
            dTics = int(round((dTop-dBottom) / dIntv))
            if (dTics <= 15):
                break # we have found a suitable scale
        return dBottom, dTop, dIntv

    # calculate scaling parameters for primary and secondary Y axis
    # returns: hBottom, hTop, hIntv, sBottom, sTop, sIntv

    def calcYScale (self, height, slope):
        if self.hscale != None:
            (hMin, hMax) = map(float, string.split(self.hscale, ','))
        else:
            hMin = float(min(height))
            hMax = float(max(height))
        if self.sscale != None:
            (sMin, sMax) = map(float, string.split(self.sscale, ','))
        else:
            # always include zero axis into slope scale
            sMin = float(min(slope+[0]))
            sMax = float(max(slope+[0]))
        # determine scale for primary axis only
        for hIntv in [5, 10, 20, 50, 100, 200, 400]:
            # round up/down
            hBottom = hIntv*int(math.floor(hMin/hIntv))
            hTop = hIntv*int(math.ceil(hMax/hIntv))
            # hBottom and hTop are integer!
            hTics = (hTop-hBottom) / hIntv
            if (hTics <= 8):
                break # we have found a suitable scale
        
        # determine scale for secondary axis only
        for sIntv in [1, 2, 4]:
            # round up/down
            sBottom = sIntv*int(math.floor(sMin/sIntv))
            sTop = sIntv*int(math.ceil(sMax/sIntv))
            # sBottom and sTop are integer!
            sTics = (sTop-sBottom) / sIntv
            if (sTics <= 8):
                break # we have found a suitable scale

        # adjust scales to match each other
        if (sTics > hTics):
            # adjust height scale
            d = sTics - hTics
            hBottom -= hIntv * (d/2)
            hTop += hIntv * (d/2)
            if d%2 == 1: # odd number
                if (hTop-hMax < hMin-hBottom):
                    hTop += hIntv
                else:
                    hBottom -= hIntv
            hTics = sTics
        elif (hTics > sTics):
            # adjust slope scale
            d = hTics - sTics
            if sBottom == 0:
                sTop += sIntv * d
            else:
                sBottom -= sIntv * (d/2)
                sTop += sIntv * (d/2)
                if d%2 == 1: # odd number
                    if (sTop-sMax < sMin-sBottom):
                        sTop += sIntv
                    else:
                        sBottom -= sIntv
            sTics = hTics
        return hBottom, hTop, hIntv, sBottom, sTop, sIntv

    # print actual graph

    def drawCurve(self, parent, x, xMin, xMax, y, yMin, yMax, color):
        # calculate derivatives for Bezier points
        x1 = self.derive(x)
        y1 = self.derive(y)
        # first point
        coords = [ self.scaleCanvas (x[0], xMin, xMax, y[0], yMin, yMax) ]
        # remaining points
        for i in range(1,len(x)):
            # bezier segment: first control point
            sx,sy = self.scaleCanvas (x[i-1] + x1[i-1]/3.0, xMin, xMax,
                                 y[i-1] + y1[i-1]/3.0, yMin, yMax)
            coords.append ((round(sx,2), round(sy,2)))
            # second control point
            sx,sy = self.scaleCanvas (x[i] - x1[i]/3.0, xMin, xMax,
                                 y[i] - y1[i]/3.0, yMin, yMax)
            coords.append ((round(sx,2), round(sy,2)))
            # end point
            sx,sy = self.scaleCanvas (x[i], xMin, xMax, y[i], yMin, yMax)
            coords.append ((round(sx,2), round(sy,2)))
        Bezier (parent, coords, style=Style(stroke=color,stroke_width=1.5,fill="none"))

    # scale coordinate such that the interval [x0,x1] is mapped
    # to [u0,u1]

    def scale (self, x, x0, x1, u0, u1):
        return u0+(u1-u0)*float(x-x0)/(x1-x0)

    # scale point to fit SVG canvas (0,0)-(500,200)

    def scaleCanvas (self, x, x0, x1, y, y0, y1):
        sx = self.scale (x, x0, x1, 0, 500)
        sy = self.scale (y, y0, y1, 200, 0)
        return sx,sy

    # approximate derivative of an array of data points
    # using quadratic interpolation between consecutive points

    def derive (self, d):
        n = len(d) # number of points
        if n <= 1:
            return d
        elif n == 2:
            return [d[1]-d[0], d[1]-d[0]]
        else:
            d1 = array.array('d') # result

            # calculate d1[0] from quadratic interpolation of d[0..2]
            d1.append((4*d[1] - 3*d[0] - d[2]) / 2.0)
            for i in range(1, n-1):
                # calculate d1[i] from adjacent points
                if (d[i+1] - d[i] >= 0 and d[i]-d[i-1] <= 0) \
                   or (d[i+1] - d[i] <= 0 and d[i]-d[i-1] >= 0):
                    di = 0.0
                else:
                    di = (d[i+1] - d[i-1]) / 2.0
                d1.append(di)
            # calculate d1[n-1] from quadratic interpolation of d[n-3..n-1]
            d1.append((3*d[n-1] + d[n-3] - 4*d[n-2]) / 2.0)
            return d1



# ----------------------------------------
# Copyright (C) 2010-2012  Frank Paehlke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import codecs
import datetime
import tourfilesmooth

try:
    import gpxdata
    _no_gpxdata = False
except ImportError:
    _no_gpxdata = True

def td2seconds (timedelta):
    return 86400 * timedelta.days + timedelta.seconds + timedelta.microseconds/1e6

# ----------------------------------------

class TourFile:
    "representation of a Tour file"
    
    # column names
    columns = ['time', 'distance', 'height', 'lat', 'lon', 'heart_rate', 'temperature',
               'smooth_time', 'smooth_distance', 'smooth_height', 'speed', 'vspeed', 'slope' ,'comment']

    # mapping from column name to column index
    _columnindex = dict([(columns[i], i) for i in range(len(columns))])
    
    def __init__ (self, title=None, date=None, copyright=None):
        self.points = []
        self.title = title
        self.date = date
        self.copyright = copyright
        self.dmin = None
        self.dmax = None
        self.hmin = None
        self.hmax = None
        self.smin = None
        self.smax = None
        self.filename = '<>'
        self.filetype= None
        self.encoding = None
        self.is_smoothed = False
        self.speed_is_original = True

    def __len__ (self):
        "return number of points"
        return len(self.points)
    
    def get (self, row, column=None):
        "get value of row-th point"
        if column == None:
            return self.points[row]
        else:
            return self.points[row][self._columnindex[column]]

    def getcolumn (self, column):
        "get all values of column"
        col = self._columnindex[column]
        return [p[col] for p in  self.points]

    def put (self, row, point=None, **kw):
        "set value(s) of row-th point"
        if point == None:
            for key, value in kw.items():
                if key == 'time':
                    value = datetime.datetime.strptime(value, '%H:%M:%S')
                elif key != 'commment':
                    value = float(value)
                self.points[row][self._columnindex[key]] = value
        elif len(point) != len(self.columns):
            raise Exception ("trying to insert point with wrong number of elements")
        else:
            self.points[row] = point

    def append (self, point=None, **kw):
        if point == None:
            # convert key/value list to point
            p = []
            for col in self.columns:
                if col in kw:
                    p.append (kw[col])
                else:
                    p.append (None)
            self.points.append (p)
        elif len(point) != len(self.columns):
            raise Exception ("trying to insert point with wrong number of elements")
        else:
            self.points.append (point)

    def delete (self, row):
        del self.points[row]

    def clear (self):
        self.points = []

    def distance (self, i, j):
        "distance between point i and j"
        col = self._columnindex['distance']
        return self.points[j][col] - self.points[i][col]

    def length (self):
        "return tour length [km]"
        if len(self.points) == 0:
            return 0.0
        else:
            col = self._columnindex['distance']
            return self.points[-1][col] - self.points[0][col]

    def hdiff (self, net=False):
        "return height difference [m]"
        if len(self.points) == 0:
            return 0.0
        else:
            col = self._columnindex['smooth_height' if self.is_smoothed else 'height']
            if net:
                return self.points[-1][col] - self.points[0][col]
            else:
                return sum([
                    self.points[i+1][col] - self.points[i][col]
                    for i in range(len(self.points)-1)
                    if self.points[i+1][col] > self.points[i][col]
                    ])

    def slope (self):
        "return average slope [%]"
        if len(self.points) == 0:
            return None
        else:
            return self.hdiff (True) / self.length () / 10.0
                
    def maxslope (self):
        "return maximum slope [%]"
        if len(self.points) == 0:
            return None
        else:
            col = self._columnindex['slope']
            return max ([ p[col] for p in self.points ])
                
    def reverse (self):
        "reverse tour direction"
        if len(self.points) == 0:
            return
        dcol = self._columnindex['distance']
        start = self.points[0][dcol]
        end = self.points[-1][dcol]
        points = []
        for i in range(len(self.points)-1, -1, -1):
            p = self.points[i]
            p[dcol] = start + end - p[dcol]
            points.append (p)
        self.points = points
        self.calc_vspeed()
        self.calc_slope()

    def heightOffset (self, offset, rows=None):
        """add offset to heights"""
        if rows == None:
            rows = range(len(self.points)) # select all rows
        hcol = self._columnindex['height']
        shcol = self._columnindex['smooth_height']
        for i in rows:
            p = self.points[i]
            p[hcol] += offset
            if p[shcol] != None:
                p[shcol] += offset
        # no need to recalculate distance/slope

    def distanceOffset (self, offset, rows=None):
        """add offset to heights"""
        if rows == None:
            rows = range(len(self.points)) # select all rows
        dcol = self._columnindex['distance']
        sdcol = self._columnindex['smooth_distance']
        for i in rows:
            p = self.points[i]
            p[dcol] = round(p[dcol] + offset, 3)
            if p[sdcol] != None:
                p[sdcol] += offset
        # no need to recalculate distance/slope

    def simplify (self, delta=1.0):
        latcol = self._columnindex['lat']
        loncol = self._columnindex['lon']
        seg = gpxdata.TrackSegment(
            [ gpxdata.TrackPoint(p[latcol], p[loncol])
              for p in self.points
              if p[latcol] != None and p[loncol] != None
              ]
            )
        n = len(seg.points)
        idx = [0] + seg._simplify (0, n-1, delta) + [n-1]
        self.points = [ self.points[i] for i in idx ]        
        self.recalc()

    def smooth (self, fixedRows=set(), tf=tourfilesmooth.TFLength,
                startHeight=None, endHeight=None):
        smooth = tourfilesmooth.TourFileSmooth (tf)
        dcol = self._columnindex['distance']
        hcol = self._columnindex['height']
        sdcol = self._columnindex['smooth_distance']
        shcol = self._columnindex['smooth_height']
        # call smoothing algorithm
        s, h = smooth.smooth (
            [p[dcol] for p in self.points],
            [p[hcol] for p in self.points],
            0.005, 0.5, fixed=fixedRows )
        if startHeight != None and endHeight != None:
            # scale height values
            h0 = h[0]
            h1 = h[-1]
            scale = (endHeight-startHeight) / (h1-h0)
            for i in range(len(h)):
                h[i] = startHeight + scale * (h[i]-h0)
        # write results to tourfile
        for i in range(len(self.points)):
            self.points[i][sdcol] = s[i]
            self.points[i][shcol] = h[i]
        self.is_smoothed = True
        self.recalc()

    def calc_speed (self):
        if self.speed_is_original or len(self.points) == 0:
            return
        tcol = self._columnindex['time']
        dcol = self._columnindex['smooth_distance' if self.is_smoothed else 'distance']
        scol = self._columnindex['speed']
        self.points[0][scol] = None
        for i in range(1, len(self.points)):
            curr = self.points[i]
            prev = self.points[i-1]
            if curr[tcol] == None or prev[tcol] == None:
                continue
            tdiff = td2seconds(curr[tcol] - prev[tcol]) / 3600.0
            distance = curr[dcol] - prev[dcol]
            if tdiff != 0:
                self.points[i][scol] = round (distance/tdiff, 2)

    def calc_vspeed (self):
        if len(self.points) == 0:
            return
        tcol = self._columnindex['time']
        hcol = self._columnindex['smooth_height' if self.is_smoothed else 'height']
        vscol = self._columnindex['vspeed']
        self.points[0][vscol] = None
        for i in range(1, len(self.points)):
            curr = self.points[i]
            prev = self.points[i-1]
            if curr[tcol] == None or prev[tcol] == None:
                continue
            tdiff = td2seconds(curr[tcol] - prev[tcol]) / 3600.0
            hdiff = curr[hcol] - prev[hcol]
            if tdiff != 0:
                self.points[i][vscol] = round (hdiff/tdiff)

    def calc_slope (self):
        if len(self.points) == 0:
            return
        dcol = self._columnindex['smooth_distance' if self.is_smoothed else 'distance']
        hcol = self._columnindex['smooth_height' if self.is_smoothed else 'height']
        slcol = self._columnindex['slope']
        self.points[0][slcol] = None
        for i in range(1, len(self.points)):
            curr = self.points[i]
            prev = self.points[i-1]
            distance = curr[dcol] - prev[dcol]
            hdiff = curr[hcol] - prev[hcol]
            if distance != 0:
                self.points[i][slcol] = round (hdiff/distance/10.0, 1)

    def recalc(self):
        self.calc_speed()
        self.calc_vspeed()
        self.calc_slope()

    @staticmethod
    def readCRP (infile, encoding='iso-8859-1'):
        if isinstance (infile, basestring):
            # infile is a file name
            fp = open (infile, "rb")
            try:
                res = TourFile.readCRP (fp)
            finally:
                fp.close()
            return res
        # infile is a file pointer
        res = TourFile()
        res.filename = infile.name
        res.filetype = 'CRP'
        res.encoding = encoding
        reader = codecs.getreader(encoding)(infile)
        line = reader.readline()
        if line.startswith("HRMProfil"):
            # input includes complete CRP header
            # we can safely ignore the 2nd header line since we do not need
            # any of the contained data
            line = reader.readline()
            # read first real input line into buffer
            line = reader.readline()
            if not line:
                return res
        while line:
            if line.startswith("***"):
                # end of data rows reached; the input seems to contain a
                # complete CRP trailer: read it
                # ignore first line
                line = reader.readline()
                if not line: return res
                # second line
                line = reader.readline()
                if not line: return res
                fields = line.rstrip().split("\t")
                res.date = datetime.datetime.strptime(fields[0]+" "+fields[1], "%d.%m.%Y %H:%M:%S")
                res.title = fields[5]
                # the rest of the file can be ignored
                return res
            # read normal data row
            fields = line.rstrip().split("\t")
            kw = {}
            kw['heart_rate'] = float(fields[0])
            kw['speed'] = float(fields[1]) / 10.0 # unit: km/h
            kw['distance'] = float(fields[2]) / 100.0 # unit: km
            kw['height'] = int(fields[3])
            kw['temperature'] = float(fields[6].replace(",","."))
            kw['time'] = datetime.datetime.strptime (fields[7], "%H:%M:%S")
            if len(fields) > 8: kw['comment'] = fields[8]
            res.append (**kw)
            line = reader.readline()
        res.is_smoothed = False
        res.speed_is_original = True
        res.recalc()
        return res

    @staticmethod
    def readCSV (infile, encoding='utf-8'):
        if isinstance (infile, basestring):
            # infile is a file name
            fp = open (infile, "rb")
            try:
                res = TourFile.readCSV (fp, encoding)
            finally:
                fp.close()
            return res
        # infile is a file object
        res = TourFile()
        res.filename = infile.name
        res.filetype = 'CSV'
        res.encoding = encoding
        reader = codecs.getreader(encoding)(infile)
        for line in reader:
            fields = line.rstrip().split('\t')
            if len(fields) == 2:
                key = fields[0].lower()
                if key == 'title':
                    res.title = fields[1]
                elif key == 'date':
                    res.date = datetime.datetime.strptime(fields[1], '%Y-%m-%d')
                elif key == 'copyright':
                    res.copyright = fields[1]
                elif key == 'dscale':
                    res.dmin, res.dmax = fields[1].split(',')
                    res.dmin = float(res.dmin)
                    res.dmax = float(res.dmax)
                elif key == 'hscale':
                    res.hmin, res.hmax = fields[1].split(',')
                    res.hmin = float(res.hmin)
                    res.hmax = float(res.hmax)
                elif key == 'sscale':
                    res.smin, res.smax = fields[1].split(',')
                    res.smin = float(res.smin)
                    res.smax = float(res.smax)
            elif len(fields) > 2 and fields[0][0] in '0123456789.+-':
                fields[0] = datetime.datetime.strptime (fields[0], "%H:%M:%S")
                for i in range (1, 10):
                    fields[i] = float(fields[i])
                res.append (
                    **dict(
                        zip(
                            ['time', 'distance', 'height', 'heart_rate',
                             'temperature', 'smooth_distance', 'smooth_height',
                             'speed', 'vspeed', 'slope', 'comment'],
                            fields)))
        res.is_smoothed = True
        res.speed_is_original = False
        res.recalc()
        return res

    @staticmethod
    def readGPX (infile):
        if _no_gpxdata:
            raise Exception("readGPX() requires module gpxdata")
        if isinstance (infile, basestring):
            # infile is a file name
            fp = open (infile, "rb")
            try:
                res = TourFile.readGPX (fp)
            finally:
                fp.close()
            return res
        # infile is a file object
        gpx = gpxdata.Document.readGPX (infile)
        # get first track
        track = gpx.tracks[0]
        # convert to Profile object
        res = TourFile (track.name, track.segments[0].points[0].t)
        res.filename = infile.name
        res.filetype = 'GPX'
        res.encoding = 'utf-8'
        s = 0.0
        for trkseg in track.segments:
            for i in range(len(trkseg.points)):
                p = trkseg.points[i]
                if i > 0:
                    s += trkseg.points[i-1].distance(p)
                res.append (
                    time=p.t, 
                    lat=p.lat,
                    lon=p.lon,
                    distance=round (s/1000.0, 3),
                    height=p.ele)
        res.is_smoothed = False
        res.speed_is_original = False
        res.recalc()
        return res

    @staticmethod
    def read (filename, encoding, filetype=None):
        if filetype == None:
            # guess filetype from filename
            fn = filename.lower()
            if fn.endswith ('.crp'):
                filetype = 'CRP'
            elif fn.endswith ('.csv'):
                filetype = 'CSV'
            elif fn.endswith ('.gpx'):
                filetype = 'GPX'
            else:
                raise Exception ("Can't determine file type")
        if filetype == 'CRP':
            return TourFile.readCRP (filename, encoding)
        elif filetype == 'CSV':
            return TourFile.readCSV (filename, encoding)
        elif filetype == 'GPX':
            return TourFile.readGPX (filename)
        else:
            raise Exception ('Unknown file type '+filetype)

    def writeCRP (self, outfile, encoding='iso-8859-1'):
        if isinstance (outfile, basestring):
            # outfile is a file name
            fp = open (outfile, "wb")
            try:
                self.writeCRP (fp)
            finally:
                fp.close()
            return
        # outfile is a file object
        self.filename = outfile.name
        self.filetype = 'CRP'
        self.encoding = encoding
        w = codecs.getwriter(encoding)(outfile)
        for i in range(len(self.points)):
            heart_rate = self.get (i, 'heart_rate')
            speed = self.get (i, 'speed')
            distance = self.get (i, 'distance')
            height = self.get (i, 'height')
            time = self.get (i, 'time')
            temperature = self.get (i, 'temperature')
            temperature = '%.1f' % (temperature if temperature != None else 0)
            temperature = temperature.replace ('.', ',')
            comment = self.get (i, 'comment')
            w.write (
                "%d\t%d\t%d\t%d\t%d\t%d\t%s\t%s\t%s\r\n" % (
                    round(heart_rate) if heart_rate != None else 0,
                    round(10*speed) if speed != None else 0,
                    round(100*distance) if distance != None else 0,
                    round(height) if height != None else 0,
                    0,
                    0,
                    temperature,
                    time.strftime ('%H:%M:%S') if time != None else '00:00:00',
                    comment if comment != None else ''
                    ))

    def writeCSV (self, outfile, encoding="utf-8"):
        if isinstance (outfile, basestring):
            # outfile is a file name
            fp = open (outfile, "wb")
            try:
                self.writeCSV (fp, encoding)
            finally:
                fp.close()
            return
        # outfile is a file object
        self.filename = outfile.name
        self.filetype = 'CSV'
        self.encoding = encoding
        w = codecs.getwriter(encoding)(outfile)
        w.write ("Encoding\t%s\n" % encoding)
        if self.title:
            w.write ("Title\t%s\n" % self.title)
        if self.date != None:
            w.write ("Date\t%s\n" % self.date.strftime("%Y-%m-%d"))
        if self.copyright:
            w.write ("Copyright\t%s\n" % self.copyright)
        if self.dmin != None and self.dmax != None:
            w.write ("Dscale\t%g,%g\n" % (self.dmin, self.dmax))
        if self.hmin != None and self.hmax != None:
            w.write ("Hscale\t%g,%g\n" % (self.hmin, self.hmax))
        if self.smin != None and self.smax != None:
            w.write ("Sscale\t%g,%g\n" % (self.smin, self.smax))
        for i in range(len(self.points)):
            w.write (
                "%s\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%s\n" % (
                    self.get (i, 'time').strftime ("%H:%M:%S"),
                    self.get (i, 'distance'),
                    self.get (i, 'height'),
                    self.get (i, 'heart_rate') or 0,
                    self.get (i, 'temperature') or 0,
                    self.get (i, 'smooth_distance') or 0,
                    self.get (i, 'smooth_height') or 0,
                    self.get (i, 'speed') or 0,
                    self.get (i, 'vspeed') or 0,
                    self.get (i, 'slope') or 0,
                    self.get (i, 'comment') or ""))

    def write (self, filename, encoding, filetype=None):
        if filetype == None:
            # guess filetype from filename
            fn = filename.lower()
            if fn.endswith ('.crp'):
                filetype = 'CRP'
            elif fn.endswith ('.csv'):
                filetype = 'CSV'
            else:
                raise Exception ("Can't determine file type")
        if filetype == 'CRP':
            self.writeCRP (filename, encoding)
        elif filetype == 'CSV':
            self.writeCSV (filename, encoding)
        else:
            raise Exception ('Unknown file type '+filetype)

# ----------------------------------------



# -*- coding: utf-8 -*-
# ----------------------------------------
# Copyright (C) 2010-2011  Frank Pählke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import array
import math
from Tkinter import *
import tkFileDialog
import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg

_figsize = (10, 6.25)
_figdpi = 96

class TourFileFigure:
    """
    matplotlib figure displaying a TourFile
    """

    def __init__ (self, tourfile):
        self.tourfile = tourfile
        self.figure = plt.figure(1, figsize=_figsize, facecolor='white', edgecolor='black', linewidth=1, frameon=True)
        self.figure.clear()
        self.draw ()

    def draw(self):
        """draw graph into Figure object self.fig"""
        fig = self.figure
        ax = fig.add_subplot(111)
        ax.patch.set_facecolor('0.9') # light grey
        tf = self.tourfile
        if tf.is_smoothed:
            dist = tf.getcolumn ('smooth_distance')
            height = tf.getcolumn ('smooth_height')
        else:
            dist = tf.getcolumn ('distance')
            height = tf.getcolumn ('height')

        ax.plot (dist, height, color='blue', zorder=2)

        # calculate slope
        dist2 = [ (a+b)/2.0 for a,b in zip(dist[0:-1], dist[1:]) ]
        slope = [ ((b-a)/(d-c)/10.0 if d != c else 0) for a,b,c,d in
                  zip(height[0:-1], height[1:], dist[0:-1], dist[1:]) ]

        ax2 = ax.twinx()
        ax2.plot (dist2, slope, color="red", zorder=3)

        ax.set_xlabel ('Entfernung [km]')
        ax.set_ylabel (u'blau: Höhe [m]')
        ax2.set_ylabel ('rot: Steigung [%]')

        # calculate scales
        d0,d1,dIntv = self.calcXScale(dist)
        h0,h1,hIntv,s0,s1,sIntv = self.calcYScale(height,slope)
        dTics = int(round((d1-d0)/dIntv))
        hTics = int(round((h1-h0)/hIntv))
        sTics = int(round((s1-s0)/sIntv))
        ax.set_xlim (d0, d1)
        ax.set_ylim (h0, h1)
        ax2.set_xlim (d0, d1)
        ax2.set_ylim (s0, s1)

        # draw grid
        ax.set_xticks ( [d0+i*dIntv for i in range(dTics+1)] )
        ax.set_yticks ( [h0+i*hIntv for i in range(hTics+1)] )
        ax2.set_xticks ( [d0+i*dIntv for i in range(dTics+1)] )
        ax2.set_yticks ( [s0+i*sIntv for i in range(sTics+1)] )
        ax.grid(True)
        if h0 < 0 and h1 > 0:
            ax.axhline (0, color='black', zorder=1)
        if s0 < 0 and s1 > 0:
            ax2.axhline (0, color='black', zorder=1)

        # title
        if tf.title != None:
            ax.set_title (tf.title)

        # copyright
        if tf.copyright != None:
            ax.text (0.99, 0.016, tf.copyright, ha='right', va='bottom',
                     transform=fig.transFigure, size='small', color='0.5')                     
        

    # calculate scaling parameters X axis
    # returns: dBottom, dTop, dIntv

    def calcXScale (self, dist):
        tf = self.tourfile
        if tf.dmin != None:
            dMin = tf.dmin
        else:
            dMin = min(dist)
        if tf.dmax != None:
            dMax = tf.dmax
        else:
            dMax = max(dist)
        # determine scale
        for dIntv in [0.1, 0.2, 0.5, 1, 2, 4, 5]:
            # round up/down
            dBottom = dIntv * math.floor(dMin/dIntv)
            dTop = dIntv * math.ceil(dMax/dIntv)
            dTics = int(round((dTop-dBottom) / dIntv))
            if (dTics <= 20):
                break # we have found a suitable scale
        return dBottom, dTop, dIntv

    # calculate scaling parameters for primary and secondary Y axis
    # returns: hBottom, hTop, hIntv, sBottom, sTop, sIntv

    def calcYScale (self, height, slope):
        tf = self.tourfile
        if tf.hmin != None:
            hMin = tf.hmin
        else:
            hMin = float(min(height))
        if tf.hmax != None:
            hMax = tf.hmax
        else:
            hMax = float(max(height))
        if tf.smin != None:
            sMin = tf.smin
        else:
            sMin = float(min(slope+[0]))
        if tf.smax != None:
            sMax = tf.smax
        else:
            # always include zero axis into slope scale
            sMax = float(max(slope+[0]))
        # determine scale for primary axis only
        for hIntv in [5, 10, 20, 50, 100, 200, 400]:
            # round up/down
            hBottom = hIntv*int(math.floor(hMin/hIntv))
            hTop = hIntv*int(math.ceil(hMax/hIntv))
            # hBottom and hTop are integer!
            hTics = (hTop-hBottom) / hIntv
            if (hTics <= 10):
                break # we have found a suitable scale
        
        # determine scale for secondary axis only
        for sIntv in [1, 2, 4]:
            # round up/down
            sBottom = sIntv*int(math.floor(sMin/sIntv))
            sTop = sIntv*int(math.ceil(sMax/sIntv))
            # sBottom and sTop are integer!
            sTics = (sTop-sBottom) / sIntv
            if (sTics <= 10):
                break # we have found a suitable scale

        # adjust scales to match each other
        if (sTics > hTics):
            # adjust height scale
            d = sTics - hTics
            hBottom -= hIntv * (d/2)
            hTop += hIntv * (d/2)
            if d%2 == 1: # odd number
                if (hTop-hMax < hMin-hBottom):
                    hTop += hIntv
                else:
                    hBottom -= hIntv
            hTics = sTics
        elif (hTics > sTics):
            # adjust slope scale
            d = hTics - sTics
            if sBottom == 0:
                sTop += sIntv * d
            else:
                sBottom -= sIntv * (d/2)
                sTop += sIntv * (d/2)
                if d%2 == 1: # odd number
                    if (sTop-sMax < sMin-sBottom):
                        sTop += sIntv
                    else:
                        sBottom -= sIntv
            sTics = hTics
        return hBottom, hTop, hIntv, sBottom, sTop, sIntv

    def save(self, filename):
        self.figure.set_size_inches(*_figsize)
        self.figure.savefig(filename, dpi=_figdpi, papertype='a4')

class NavigationToolbar(NavigationToolbar2TkAgg):

    def save_figure(self):
        # reset figsize to original value
        self.canvas.figure.set_size_inches(*_figsize)
        matplotlib.rcParams['savefig.dpi'] = _figdpi
        NavigationToolbar2TkAgg.save_figure(self)
        self.canvas.draw()

class TourGraph(Toplevel):
    """
    Tkinter frontend for TourFileFigure
    """

    def __init__ (self, parent):
        Toplevel.__init__ (self)
        self.parent = parent
        parent.registerListener (self)
        self.fig = TourFileFigure(self.parent.tourfile)

        self.zoom = 1.0
        self.buttons = {}

        # button frame
        bf = Frame (self);
        bf.pack (side=TOP)

        self.buttons['close'] = b = Button (bf, text='Close', command=self.close)
        b.pack (side=LEFT)

        self.canvasframe = cf = Frame (self)
        cf.pack(side=TOP, fill=BOTH, expand=1)

        self.canvas = c = FigureCanvasTkAgg(self.fig.figure, cf)
        c.show()
        c.get_tk_widget().pack(side=TOP, fill=BOTH, expand=1)

        self.toolbar = tb = NavigationToolbar(c, cf)
        tb.update()
        c._tkcanvas.pack(side=TOP, fill=BOTH, expand=1)

        self.fig.draw()
        self.canvas.draw()

    def close(self):
        self.parent.deregisterListener (self)
        self.destroy()



# -*- coding: utf-8 -*-
# ----------------------------------------
# Copyright (C) 2010-2012  Frank Pählke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import sys
import datetime
from Tkinter import *
import tkFileDialog
import tkMessageBox
import tkSimpleDialog
if sys.version_info[0] >= 3:
    import tkinter.ttk as ttk
else:
    import ttk
from toureditorlib.tourfile import TourFile
from toureditorlib import tourfilesmooth

# Is the matplotlib version of TourGraph available?
try:
    from toureditorlib.tourgraph_matplotlib import TourGraph
except ImportError:
    # use standard version without matplotlib
    from toureditorlib.tourgraph import TourGraph

from toureditorlib._version import __version__

# ----------------------------------------

_encodings = [
    'iso-8859-1',
    'iso-8859-2',
    'iso-8859-3',
    'iso-8859-4',
    'iso-8859-5',
    'iso-8859-6',
    'iso-8859-7',
    'iso-8859-8',
    'iso-8859-9',
    'iso-8859-10',
    'iso-8859-11',
    'iso-8859-12',
    'iso-8859-13',
    'iso-8859-14',
    'iso-8859-15',
    'iso-8859-16',
    'utf-8'
    ]

_filetypes = [
    ('CRP', '*.crp'),
    ('CSV', '*.csv'),
    ('GPX', '*.gpx'),
    ('all', '*')
    ]

# ----------------------------------------

class TourView (Toplevel):

    def __init__ (self, tourfile):

        Toplevel.__init__ (self)
        self.tourfile = tourfile
        self.listeners = []
        self.fixedRows = set()

        self.title('Tour Editor - '+tourfile.filename)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=1)

        # 1st button frame
        bf = ttk.Frame (self)
        bf.grid (row=0, column=0, columnspan=2, sticky=W)

        l = ttk.Label (bf, text='Encoding:')
        l.pack (side=LEFT)

        self.encodingBox = ttk.Combobox(bf, values=_encodings)
        self.encodingBox.set (tourfile.encoding or 'iso-8859-1')
        self.encodingBox.pack (side=LEFT, fill=Y, expand=True)

        self.buttons = {}
        self.buttons['open'] = b = ttk.Button (bf, text='Open', command=self.open)
        b.pack(side=LEFT)

        self.buttons['save'] = b = ttk.Button (bf, text='Save', command=self.save, state=DISABLED)
        b.pack(side=LEFT)

        self.buttons['saveAs'] = b = ttk.Button (bf, text='Save As', command=self.saveAs)
        b.pack(side=LEFT)

        self.buttons['reload'] = b = ttk.Button (bf, text='Reload', command=self.reload, state=DISABLED)
        b.pack(side=LEFT)

        self.buttons['close'] = b = ttk.Button (bf, text='Close', command=self.close)
        b.pack(side=LEFT)

        # Treeview with scrollbars
        t = ttk.Treeview (self, columns=tourfile.columns)
        t.grid(row=1,column=0, sticky=NSEW)
        self.treeview = t
        t.column ('#0', width=60, stretch=False)
        for i in range(len(tourfile.columns)-1):
            t.column (i, width=100, stretch=False)
        t.tag_configure ('standstill', background='yellow')
        t.tag_configure ('fixed', foreground='red')

        self.vscr = ttk.Scrollbar (self, command=self.treeview.yview)
        t.config (yscrollcommand=self.vscr.set)
        self.vscr.grid(row=1,column=1, sticky=NS)

        self.hscr = ttk.Scrollbar (self, orient=HORIZONTAL, command=self.treeview.xview)
        t.config (xscrollcommand=self.hscr.set)
        self.hscr.grid(row=2,column=0, sticky=EW)

        t.bind ('<Double-Button-1>', self.doubleclick)
        self.bind ('<Control-a>', self.selectAll)
        self.bind ('<Control-n>', self.selectNone)

        # 2nd button frame
        bf = ttk.Frame (self)
        bf.grid (row=3, column=0, columnspan=2)

        self.buttons['meta'] = b = ttk.Button (bf, text='Metadata', command=self.metadata)
        b.pack(side=LEFT)

        self.buttons['hOffset'] = b = ttk.Button (bf, text='Height Offset', command=self.heightOffset)
        b.pack(side=LEFT)

        self.buttons['dOffset'] = b = ttk.Button (bf, text='Distance Offset', command=self.distanceOffset)
        b.pack(side=LEFT)

        self.buttons['delete'] = b = ttk.Button (bf, text='Delete Rows', command=self.delete)
        b.pack(side=LEFT)

        self.buttons['reverse'] = b = ttk.Button (bf, text='Reverse File', command=self.reverse)
        b.pack(side=LEFT)

        self.buttons['fix'] = b = ttk.Button (bf, text='Fix rows', command=self.fix)
        b.pack(side=LEFT)

        self.buttons['simplify'] = b = ttk.Button (bf, text='Simplify', command=self.simplify)
        b.pack(side=LEFT)

        self.buttons['smooth'] = b = ttk.Button (bf, text='Smooth', command=self.smooth)
        b.pack(side=LEFT)

        self.buttons['graph'] = b = ttk.Button (bf, text='Show Graph', command=self.graph)
        b.pack(side=LEFT)

        self.geometry ('800x600')
        self.refresh()

    def registerListener (self, listener):
        self.listeners.append (listener)

    def deregisterListener (self, listener):
        self.listeners = [l for l in self.listeners if l != listener]

    def dirty (self, isdirty):
        if isdirty:
            self.buttons['save'].config (state=NORMAL)
            self.buttons['reload'].config (state=NORMAL)
        else:
            self.buttons['save'].config (state=DISABLED)
            self.buttons['reload'].config (state=DISABLED)

    def refresh (self):
        t = self.treeview
        items = t.get_children()
        if len(items) > 0:
            t.delete (*items)
        tf = self.tourfile
        for i in range(len(tf.columns)):
            t.heading (i, text=tf.columns[i])
        for i in range(len(tf)):
            point = tf.get(i)[:]
            for j in range (len(point)):
                if point[j] == None:
                    point[j] = ''
                elif isinstance(point[j], datetime.datetime):
                    point[j] = point[j].strftime ('%H:%M:%S')
                elif isinstance(point[j], float):
                    point[j] = round(point[j], 3)
            tags = []
            if i > 0 and abs(tf.distance (i-1, i)) < .001:
                tags.append ('standstill')
            if i in self.fixedRows:
                tags.append ('fixed')
            t.insert ('', 'end', text=str(i+1), value=point, tags=tuple(tags))
        for l in self.listeners:
            l.refresh()

    def open(self):
        filename = tkFileDialog.askopenfilename (
            title='Open file', filetypes=_filetypes)
        if filename != None and filename != '':
            encoding = self.encodingBox.get()
            TourView (TourFile.read (filename, encoding))

    def save(self):
        tf = self.tourfile
        if tkMessageBox.askokcancel (title='Save Tour File', message='Really overwrite contents of '+tf.filename+'?'):
            tf.write (tf.filename, self.encodingBox.get(), tf.filetype)
            self.dirty (False)

    def saveAs(self):
        filename = tkFileDialog.asksaveasfilename (
            title='Save File', filetypes=_filetypes)
        if filename != None and filename != '':
            self.tourfile.write (filename, self.encodingBox.get())
            self.title (filename)
            self.dirty (False)

    def reload(self):
        tf = self.tourfile
        self.tourfile = TourFile.read (tf.filename, tf.encoding, tf.filetype)
        self.refresh()
        self.dirty (False)

    def close(self):
        for l in self.listeners:
            l.close()
        self.destroy()

    def delete (self):
        t = self.treeview
        tf = self.tourfile
        rows = [t.index(item) for item in t.selection()]
        rows.sort(reverse=True)
        for row in rows:
            tf.delete (row)
        self.refresh()
        self.dirty (True)

    def fix (self):
        t = self.treeview
        self.fixedRows = self.fixedRows.symmetric_difference(
            [t.index(item) for item in t.selection()])
        self.refresh()

    def doubleclick (self, event):
        t = self.treeview
        item = t.identify_row (event.y)
        col = t.identify_column (event.x)
        if item != '' and col != '#0':
            s = tkSimpleDialog.askstring (title='Enter value', prompt=t.heading(col, option='text'),
                                          initialvalue=t.set(item, col))
            t.set (item, col, s)
            self.tourfile.set (t.index(item), **{t.column(col, 'id'): s})
            self.dirty (True)
            for l in self.listeners:
                l.refresh()

    def selectAll (self, event):
        t = self.treeview
        t.selection_set (t.get_children())

    def selectNone (self, event):
        self.treeview.selection_set(())

    def metadata (self):
        metadataDialog = MetadataDialog(self)
        self.dirty (True)

    def heightOffset (self):
        t = self.treeview
        tf = self.tourfile
        offset = tkSimpleDialog.askinteger(title='Height Offset', prompt='Select offset [m] to add/subtract from height')
        if offset != None:
            rows = [ t.index(item) for item in t.selection() ]
            tf.heightOffset (offset, rows)
            self.refresh()
            self.dirty(True)

    def distanceOffset (self):
        t = self.treeview
        tf = self.tourfile
        offset = tkSimpleDialog.askfloat(title='Distance Offset', prompt='Select offset [km] to add/subtract from distance')
        if offset != None:
            rows = [ t.index(item) for item in t.selection() ]
            tf.distanceOffset (offset, rows)
            self.refresh()
            self.dirty(True)

    def reverse (self):
        self.tourfile.reverse()
        self.refresh()
        self.dirty(True)

    def simplify(self):
        self.tourfile.simplify(5.0)
        self.refresh()
        self.dirty(True)

    def smooth (self):
        smoothDialog = SmoothDialog(self)

    def graph(self):
        tf = self.tourfile
        TourGraph (self)

# ----------------------------------------

class MetadataDialog (tkSimpleDialog.Dialog):

    def body(self, master):
        self.title('Metadata')

        tf = self.parent.tourfile

        ttk.Label (master, text='Title ').grid (row=0, column=0, sticky=E)
        self.tourtitle = ttk.Entry (master, width=40)
        if tf.title != None:
            self.tourtitle.insert ('0', tf.title)
        self.tourtitle.grid (row=0, column=1, sticky=W)

        ttk.Label (master, text='Date ').grid (row=1, column=0, sticky=E)
        self.date = ttk.Entry (master, width=10)
        if tf.date != None:
            self.date.insert ('0', tf.date.strftime("%Y-%m-%d"))
        self.date.grid (row=1, column=1, sticky=W)

        ttk.Label (master, text='Copyright ').grid (row=2, column=0, sticky=E)
        self.copyright = ttk.Entry (master, width=40)
        if tf.copyright != None:
            self.copyright.insert ('0', tf.copyright)
        self.copyright.grid (row=2, column=1, sticky=W)

        ttk.Label (master, text='Length [km] ').grid (row=3, column=0, sticky=E)
        ttk.Label (master, text=tf.length(), relief=SUNKEN).grid (row=3, column=1, sticky=W)

        ttk.Label (master, text='Total height difference [m] ').grid (row=4, column=0, sticky=E)
        ttk.Label (master, text=int(round(tf.hdiff(False))), relief=SUNKEN).grid (row=4, column=1, sticky=W)

        ttk.Label (master, text='Net height difference [m] ').grid (row=5, column=0, sticky=E)
        ttk.Label (master, text=int(round(tf.hdiff(True))), relief=SUNKEN).grid (row=5, column=1, sticky=W)

        ttk.Label (master, text='Average slope [%] ').grid (row=6, column=0, sticky=E)
        ttk.Label (master, text=round(tf.slope(), 2), relief=SUNKEN).grid (row=6, column=1, sticky=W)

        ttk.Label (master, text='Maximum slope [%] ').grid (row=7, column=0, sticky=E)
        ttk.Label (master, text=round(tf.maxslope(), 2), relief=SUNKEN).grid (row=7, column=1, sticky=W)

        ttk.Label (master, text='Graph d_min [%] ').grid (row=8, column=0, sticky=E)
        self.dmin = ttk.Entry (master, width=10)
        if tf.dmin != None:
            self.dmin.insert ('0', str(tf.dmin))
        self.dmin.grid (row=8, column=1, sticky=W)

        ttk.Label (master, text='Graph d_max [%] ').grid (row=9, column=0, sticky=E)
        self.dmax = ttk.Entry (master, width=10)
        if tf.dmax != None:
            self.dmax.insert ('0', str(tf.dmax))
        self.dmax.grid (row=9, column=1, sticky=W)

        ttk.Label (master, text='Graph h_min [%] ').grid (row=10, column=0, sticky=E)
        self.hmin = ttk.Entry (master, width=10)
        if tf.hmin != None:
            self.hmin.insert ('0', str(tf.hmin))
        self.hmin.grid (row=10, column=1, sticky=W)

        ttk.Label (master, text='Graph h_max [%] ').grid (row=11, column=0, sticky=E)
        self.hmax = ttk.Entry (master, width=10)
        if tf.hmax != None:
            self.hmax.insert ('0', str(tf.hmax))
        self.hmax.grid (row=11, column=1, sticky=W)

        ttk.Label (master, text='Graph s_min [%] ').grid (row=12, column=0, sticky=E)
        self.smin = ttk.Entry (master, width=10)
        if tf.smin != None:
            self.smin.insert ('0', str(tf.smin))
        self.smin.grid (row=12, column=1, sticky=W)

        ttk.Label (master, text='Graph s_max [%] ').grid (row=13, column=0, sticky=E)
        self.smax = ttk.Entry (master, width=10)
        if tf.smax != None:
            self.smax.insert ('0', str(tf.smax))
        self.smax.grid (row=13, column=1, sticky=W)

        return None

    def apply(self):
        tf = self.parent.tourfile
        tf.title = self.tourtitle.get()
        date = self.date.get()
        if date != None and date != '':
            tf.date = datetime.datetime.strptime (date, "%Y-%m-%d")
        tf.copyright = self.copyright.get()
        dmin = self.dmin.get()
        try:
            tf.dmin = (float(dmin) if dmin != '' else None)
        except ValueError:
            pass
        dmax = self.dmax.get()
        try:
            tf.dmax = (float(dmax) if dmax != '' else None)
        except ValueError:
            pass
        hmin = self.hmin.get()
        try:
            tf.hmin = (float(hmin) if hmin != '' else None)
        except ValueError:
            pass
        hmax = self.hmax.get()
        try:
            tf.hmax = (float(hmax) if hmax != '' else None)
        except ValueError:
            pass
        smin = self.smin.get()
        try:
            tf.smin = (float(smin) if smin != '' else None)
        except ValueError:
            pass
        smax = self.smax.get()
        try:
            tf.smax = (float(smax) if smax != '' else None)
        except ValueError:
            pass
        
# ----------------------------------------

class SmoothDialog (tkSimpleDialog.Dialog):

    _smoothMethods = (
        ('Curve length', tourfilesmooth.TFLength),
        ('2nd derivative', tourfilesmooth.TFD2Square)
        )

    def body(self, master):
        self.title('Smoothing')

        tf = self.parent.tourfile

        self.scaleFlag = IntVar(value=0)
        cb = ttk.Checkbutton (master, text='Scale height', variable=self.scaleFlag, command=self.cbScaleFlag)
        cb.grid(row=1, column=0)

        self.startHeight = IntVar(value=tf.get(0, 'height'))
        ttk.Label (master, text='Start height:').grid (row=2, column=0)
        self.entryStartHeight = e = ttk.Entry (master, textvariable=self.startHeight, state=DISABLED )
        e.grid (row=2, column=1)

        self.endHeight = IntVar(value=tf.get(-1, 'height'))
        ttk.Label (master, text='End height:').grid (row=3, column=0)
        self.entryEndHeight = e = ttk.Entry (master, textvariable=self.endHeight, state=DISABLED)
        e.grid (row=3, column=1)

        ttk.Label (master, text='Smooth method:').grid (row=4, column=0)
        self.cbSmoothMethod = cb = ttk.Combobox (
            master, state='readonly', values=tuple([sm[0] for sm in self._smoothMethods]))
        cb.set (self._smoothMethods[0][0])
        cb.grid (row=4, column=1)
        return None

    def cbScaleFlag(self):
        if self.scaleFlag.get():
            state=NORMAL
        else:
            state=DISABLED
        self.entryStartHeight.config (state=state)
        self.entryEndHeight.config (state=state)

    def apply(self):
        scaleFlag = bool(self.scaleFlag.get())
        if scaleFlag:
            startHeight = int(self.entryStartHeight.get())
            endHeight = int(self.entryEndHeight.get())
        smDict = dict(self._smoothMethods)
        tf = smDict[self.cbSmoothMethod.get()]
        # do it!
        p = self.parent
        if scaleFlag:
            p.tourfile.smooth (p.fixedRows, tf, startHeight, endHeight)
        else:
            p.tourfile.smooth (p.fixedRows, tf)
        p.refresh()
        p.dirty(True)

# ----------------------------------------

class AboutBox (Toplevel):

    def __init__(self, parent):
        Toplevel.__init__(self, parent)
        self.title('About TourEditor')
        self.transient(parent)
        ttk.Label(self, text='TourEditor '+__version__).grid (row=0, column=0)
        ttk.Label(self, text=u'© 2010-2012 Frank Pählke <frank@kette-links.de>').grid (row=1, column=0)
        ttk.Label(self, borderwidth=2, relief=SUNKEN,
                  text="""TourEditor is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TourEditor is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.""").grid (row=2, column=0)
        ttk.Button(self, text='OK', command=self.destroy).grid(row=3, column=0)
        self.grab_set()
        self.wait_window()

# ----------------------------------------

class TourEditor (Tk):
    "main application window"

    def __init__ (self):
        Tk.__init__ (self)
        self.title ('Tour Editor')
        l = ttk.Label (text='Encoding:')
        l.pack (side=LEFT)
        c = ttk.Combobox(values=_encodings)
        c.set ('iso-8859-1')
        c.pack (side=LEFT)
        self.encodingBox = c
        b = ttk.Button (text='Open File', command=self.open)
        b.pack (side=LEFT)
        b = ttk.Button (text='About', command=self.about)
        b.pack (side=LEFT)
        b = ttk.Button (text='Exit', command=self.quit)
        b.pack (side=LEFT)

    def open (self):
        filename = tkFileDialog.askopenfilename (
            title='Open File', filetypes=_filetypes)
        if filename != None and filename != '':
            encoding = self.encodingBox.get()
            TourView (TourFile.read (filename, encoding))

    def about (self):
        aboutBox = AboutBox(self)

# ----------------------------------------



# ----------------------------------------
# Copyright (C) 2010-2011  Frank Paehlke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import math

# ----------------------------------------

class TFLength:
    "Target function: Total length of graph"

    # L = sqrt(x^2+y^2)
    # dL/dx = 2x / 2 L = x/L
    # d2L/dx2 = (1 * L - x * x/L) / L^2 = (L^2-x^2)/L^3 = y^2/L^3

    @staticmethod
    def f (X, Y):
        assert len(X) == len(Y)
        n = len(X)-1
        res = 0.0
        for i in range(n):
            dx = X[i+1] - X[i]
            dy = Y[i+1] - Y[i]
            res += math.hypot (dx, dy)
        return res

    @staticmethod
    def dfdX (X, Y, i):
        "df(X,Y)/dX[i]"
        assert len(X) == len(Y)
        n = len(X)-1
        res = 0.0
        if i > 0:
            dx = X[i] - X[i-1]
            dy = Y[i] - Y[i-1]
            res += dx / math.hypot (dx, dy)
        if i < n:
            dx = X[i] - X[i+1]
            dy = Y[i] - Y[i+1]
            res += dx / math.hypot (dx, dy)
        return res

    @staticmethod
    def dfdY (X, Y, i):
        "df(X,Y)/dY[i]"
        assert len(X) == len(Y)
        n = len(X)-1
        res = 0.0
        if i > 0:
            dx = X[i] - X[i-1]
            dy = Y[i] - Y[i-1]
            res += dy / math.hypot (dx, dy)
        if i < n:
            dx = X[i] - X[i+1]
            dy = Y[i] - Y[i+1]
            res += dy / math.hypot (dx, dy)
        return res

    @staticmethod
    def d2fdXX (X, Y, i):
        "d^2f(X,Y)/dX[i]dX[i]"
        assert len(X) == len(Y)
        n = len(X)-1
        res = 0.0
        if i > 0:
            dx = X[i] - X[i-1]
            dy = Y[i] - Y[i-1]
            h = math.hypot (dx, dy)
            res += dy*dy / (h*h*h)
        if i < n:
            dx = X[i] - X[i+1]
            dy = Y[i] - Y[i+1]
            h = math.hypot (dx, dy)
            res += dy*dy / (h*h*h)
        return res

    @staticmethod
    def d2fdYY (X, Y, i):
        "d^2f(X,Y)/dY[i]dY[i]"
        assert len(X) == len(Y)
        n = len(X)-1
        res = 0.0
        if i > 0:
            dx = X[i] - X[i-1]
            dy = Y[i] - Y[i-1]
            h = math.hypot (dx, dy)
            res += dx*dx / (h*h*h)
        if i < n:
            dx = X[i] - X[i+1]
            dy = Y[i] - Y[i+1]
            h = math.hypot (dx, dy)
            res += dx*dx / (h*h*h)
        return res

# ----------------------------------------

class TFD2Square:
    "Target function: Square sum of 2nd derivative"

    @staticmethod
    def f (X, Y):
        """
The target function:
tf(X,Y) = sum ((2*X[i] - X[i-1] - X[i+1])^2, i=1..n-1)
        + sum ((2*Y[i] - Y[i-1] - Y[i+1])^2, i=1..n-1)
"""
        n = len(X)-1
        res = 0.0
        for i in range(1, n):
            a = 2*X[i] - X[i-1] - X[i+1]
            res += a*a
            a = 2*Y[i] - Y[i-1] - Y[i+1]
            res += a*a
        return res;

    @staticmethod
    def dfdX (X, Y, i):
        "df(X,Y)/dX[i]"
        n = len(X)-1
        if i < 0 or i > n:
            return None
        elif n < 2:
            return 0
        elif i == 0:
            return X[0] - 2*X[1] + X[2]
        elif i == 1:
            return -2*X[0] + 5*X[1] - 4*X[2] + X[3]
        elif i == n-1:
            return X[i-2] - 4*X[i-1] + 5*X[i] - 2*X[i+1]
        elif i == n:
            return X[i-2] - 2*X[i-1] + X[i]
        else:
            return X[i-2] - 4*X[i-1] + 6*X[i] - 4*X[i+1] + X[i+2]

    @staticmethod
    def dfdY (X, Y, i):
        "df(X,Y)/dY[i]"
        n = len(Y)-1
        if i < 0 or i > n:
            return None
        elif n < 2:
            return 0
        elif i == 0:
            return Y[0] - 2*Y[1] + Y[2]
        elif i == 1:
            return -2*Y[0] + 5*Y[1] - 4*Y[2] + Y[3]
        elif i == n-1:
            return Y[i-2] - 4*Y[i-1] + 5*Y[i] - 2*Y[i+1]
        elif i == n:
            return Y[i-2] - 2*Y[i-1] + Y[i]
        else:
            return Y[i-2] - 4*Y[i-1] + 6*Y[i] - 4*Y[i+1] + Y[i+2]

    @staticmethod
    def d2fdXX (X, Y, i):
        "d^2f(X,Y)/dX[i]dX[i]"
        n = len(X) - 1
        if i < 0 or i > n:
            return None
        if n < 2:
            return 0
        elif i == 0 or i == n:
            return 1
        elif i == 1 or i == n-1:
            return 5
        else:
            return 6

    @staticmethod
    def d2fdYY (X, Y, i):
        "d^2f(X,Y)/dY[i]dY[i]"
        n = len(Y) - 1
        if i < 0 or i > n:
            return None
        if n < 2:
            return 0
        elif i == 0 or i == n:
            return 1
        elif i == 1 or i == n-1:
            return 5
        else:
            return 6

# ----------------------------------------

class TourFileSmooth:

    def __init__ (self, targetfunction):
        self.f = targetfunction.f
        self.dfdX = targetfunction.dfdX
        self.dfdY = targetfunction.dfdY
        self.d2fdXX = targetfunction.d2fdXX
        self.d2fdYY = targetfunction.d2fdYY

    def smooth (self, X, Y, maxdeltaX, maxdeltaY, accuracy=1e-5, fixed=set()):
        """
minimize f(X) under the condition (res[i]-X[i] <= maxdelta)

tf(X) is minimized using an iterative algorithm until the
maximum change to any X[i] between two subsequent iterations
is smaller than maxdelta*accuracy
"""
        assert len(X) == len(Y)
        n = len(X) - 1
        if n < 2:
            return X, Y

        # normalize X, Y such that maxdelta=1
        X = [x / maxdeltaX for x in X]
        Y = [y / maxdeltaY for y in Y]
        # make a copy of X and Y
        resX = X[:]
        resY = Y[:]
        #print [('%.5f' % x) for x in resX], [('%.5f' % y) for y in resY]

        maxcorr = float('inf') # maximum correction
        while maxcorr >= accuracy:
            maxcorr = 0.0
            for i in range(1, n):
                if i in fixed:
                    continue
                # correct X[i]
                df = self.dfdX (resX, resY, i)
                d2f = self.d2fdXX (resX, resY, i)
                # correction term
                if df == 0.0 or d2f == 0.0:
                    continue
                resX_i = resX[i] - df/d2f
                # border condition
                if resX_i < X[i] - 1:
                    resX_i = X[i] - 1
                elif resX_i > X[i] + 1:
                    resX_i = X[i] + 1
                # maximum correction?
                corr = abs (resX_i - resX[i])
                if corr > maxcorr:
                    maxcorr = corr
                resX[i] = resX_i
                # correct Y[i]
                df = self.dfdY (resX, resY, i)
                d2f = self.d2fdYY (resX, resY, i)
                # correction term
                resY_i = resY[i] - df/d2f
                # border condition
                if resY_i < Y[i] - 1:
                    resY_i = Y[i] - 1
                elif resY_i > Y[i] + 1:
                    resY_i = Y[i] + 1
                # maximum correction?
                corr = abs (resY_i - resY[i])
                if corr > maxcorr:
                    maxcorr = corr
                resY[i] = resY_i
            #print [('%.5f' % x) for x in resX], [('%.5f' % y) for y in resY], maxcorr, accuracy

        # de-normalize X, Y
        return [x * maxdeltaX for x in resX], [y * maxdeltaY for y in resY]

# ----------------------------------------



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from distutils.core import setup

# read version number into __version__
exec(open('toureditorlib/_version.py').read())

setup(name='toureditor',
      version=__version__,
      description=(
        'Editor for GPX/CRP tour files and height profiles.\n'
        '\n'
        'Generates nicer output if matplotlib is installed.'),
      author='Frank Pählke',
      author_email='frank@kette-links.de',
      url='http://www.kette-links.de/technik/',
      download_url='http://www.kette-links.de/software/toureditor-'+__version__+'.tar.gz',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: End Users/Desktop',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering :: GIS',
          ],
      packages=['toureditorlib'],
      scripts=['toureditor.pyw', 'tourgraph.py'],
      requires=['mx.DateTime','pyttk','gpxdata','svgobject']
     )



#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-
# ----------------------------------------
# Copyright (C) 2010-2011  Frank Pählke
#
# This file is part of TourEditor.
#
# TourEditor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TourEditor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TourEditor.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------

import sys
from toureditorlib.tourfile import TourFile
from toureditorlib.tourgraph_matplotlib import TourFileFigure

encoding = 'iso-8859-1'
infilename = sys.argv[1]
outfilename = sys.argv[2]

sys.stdout.write ('Converting %s to %s\n' % (infilename, outfilename))

tourfile = TourFile.read (infilename, encoding)
fig = TourFileFigure (tourfile)
fig.save(outfilename)
